(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.


(*************************************************************************)
(* Manifest types *)
(*************************************************************************)

(* [=t:k]   :=  (\a. !b:(k->O).b a -> b a) t*)
Definition typT t k : typ :=
  appT (lamT k (allT (arrK k omK) (arrT (appT 0 1) (appT 0 1)))) t.

(* [t:k]   :=   (\a. \b:(k->O).\x:(b a).x) [t] *)
Definition typE t k : exp :=
 instE (genE k (genE (arrK k omK) (lamE (appT 0 1) 0))) t.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substAT_typ : forall a t' t k,
  lclosedT t' ->
  lclosedT t ->
  substAT a t' (typT t k) = typT (substAT a t' t) k.
Proof.
  intros a t' t k Ht' Ht. unfold typT.
  rewrite substAT_app.
  destruct a.
  reflexivity.
  rewrite substAT_lclosed.  
  reflexivity.
  eapply lclosedT_lam with (L := {}).
  intros. 
  unfold openAT.
  simpl.
  eapply lclosedT_all with (L := {}).
  intros.
  unfold openAT.
  simpl.
  auto*.
Qed.

Corollary openAT_typ : forall t' t k,
  lclosedT t' ->
  lclosedT t ->
  openAT t' (typT t k) = typT (openAT t' t) k.
Proof.
  intros t' t k Ht' Ht. unfold openAT.
  apply substAT_typ. apply Ht'. apply Ht.
Qed.


Lemma substAE_typ : forall a t' t k,
  lclosedT t' ->
  lclosedT t ->
  substAE a t' (typE t k) = typE (substAT a t' t) k.
Proof.
  intros a t' t k Ht' Ht. unfold typE.
  unfold substAE.
  destruct a; simpl shift.
    reflexivity.
    rewrite (substAT_lclosed n t' t). 
    simpl.
    reflexivity.
    trivial.
Qed.

Corollary openAE_typ : forall t' t k,
  lclosedT t' ->
  lclosedT t ->
  openAE t' (typE t k) = typE (openAT t' t) k.
Proof.
  intros t' t k Ht' Ht. unfold openAE, openAT.
  apply substAE_typ. apply Ht'. apply Ht.
Qed.


Lemma substXE_typ : forall x e' t k,
  lclosedE e' ->
  lclosedT t ->
  substXE x e' (typE t k) = typE t k.
Proof.
  intros x e' t k He' Ht. unfold typE.
    destruct x; simpl; reflexivity.
Qed.

Corollary openXE_typ : forall e' t k,
  lclosedE e' ->
  lclosedT t ->
  openXE e' (typE t k) = typE t k.
Proof.
  intros e' t k He' Ht. unfold openXE.
  apply substXE_typ. apply He'. apply Ht.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Lemma lclosedT_typ : forall t k,
  lclosedT t ->
  lclosedT (typT t k).
Proof.
  intros t k Ht. unfold typT.
  apply lclosedT_app.
   apply lclosedT_lam with (L := {}).
   intros a Fr.
   unfold openAT; simpl.
   apply lclosedT_all with (L := {}).
   intros b Frb.
   unfold openAT; simpl. 
   auto*.
  trivial.
Qed.

Lemma lclosedT_typ_inversion : forall t k,
  lclosedT (typT t k) ->
  lclosedT t.
Proof.
  intros t k Ht. unfold typT in Ht.
  inversion Ht.
  trivial.
Qed.

Lemma lclosedE_typ : forall t k,
  lclosedT t ->
  lclosedE (typE t k).
Proof.
  intros t k Ht. unfold typE.
  apply lclosedE_inst.
    apply lclosedE_gen with (L := {}). intros a Fr.
      unfold openAE. simpl substAE.
         apply lclosedE_gen with (L := {}). intros b FrB. 
         unfold openAE.  simpl substAE. 
         apply lclosedE_lam with (L := {}).
         apply lclosedT_app; apply lclosedT_var.
        intros x Fr'. unfold openXE. simpl substXE. apply lclosedE_var.
    apply Ht.
Qed.


(*************************************************************************)
(* Typing*)
(*************************************************************************)

Lemma ofT_typ : forall D t k,
  ofT D t k ->
  ofT D (typT t k) omK.
Proof.
  intros D t k Ht. unfold typT.
  eapply ofT_app.
  2: apply Ht.
  apply ofT_lam with (L := dom D). intros b Frb.
    unfold openAT. simpl substAT.
    apply ofT_all with (L := dom (D & b ~ k)). intros a Fr.
    unfold openAT.
    simpl.
    assert (HD'' : okD (D & b ~ k )).
     apply okD_var.  eapply validityDT. apply Ht. apply Frb. 
    assert (HD' : okD (D & b ~ k & a ~ arrK k omK)).
      apply okD_var.  
      apply HD''. apply Fr.
    assert (Happ : ofT (D & b ~ k & a ~ arrK k omK) (appT a b) omK).
      eapply ofT_app.
        apply ofT_var. apply HD'. apply binds_head.
          apply weakeningDT_head. apply ofT_var.  apply HD''.
          apply binds_head.
    apply HD'.
    apply ofT_arr. apply Happ. apply Happ.
Qed.

Lemma ofT_typ_inversion : forall D t k,
  lclosedT t ->
  ofT D (typT t k) omK ->
  ofT D t k.
Proof.
  intros D t k Ht H.
  unfold typT in H.
  inversion H; subst.
  assert (RW : k = k2).
   inversion H3.
   reflexivity.
  rewrite RW.
  trivial.
Qed.   
 
Lemma eqT_typ: forall t t' k,
  lclosedT t ->
  lclosedT t' ->
  eqT t t' ->
  eqT (typT t k) (typT t' k).
Proof.
  intros t t' k Ht Ht' H.
  unfold typT.
  apply eqT_app.
  apply eqT_refl.
  apply lclosedT_lam with (L := {}).
   intros a Fr.
   unfold openAT; simpl.
   apply lclosedT_all with (L := {}).
   intros b Frb.
   unfold openAT; simpl. 
   auto*.
  apply H.
Qed.


Lemma ofE_typ : forall D G t k,
  okG D G ->
  ofT D t k ->
  ofE D G (typE t k) (typT t k).
Proof.
  intros D G t k HG Ht.
  unfold typE, typT.
  eapply ofE_sub.
  Focus 3.
   apply eqT_sym.
   apply eqT_app_beta.
     eapply lclosedT_lam with (L := {}).
     intros. 
     unfold openAT.
     simpl.
     eapply lclosedT_all with (L := {}).
     intros.
     unfold openAT.
     simpl.
     auto*.
    apply (regularityT  _ _ _ Ht).
  eapply ofE_inst.
  apply ofE_gen with (L := dom D). apply HG. intros a Fra.
    unfold openAE, openAT. simpl substAE. simpl substAT.
    assert (HD' : okD (D & a ~ arrK k omK)).
      apply okD_var. eapply validityDT. apply Ht. apply Fra.
    assert (HG' : okG (D & a ~ arrK k omK) G).
      apply weakeningDG_head. apply HG. apply HD'.
    assert (Happ : ofT (D & a ~ arrK k omK) (appT a (substAT 0 a t)) omK).
       eapply ofT_app.
         apply ofT_var. apply HD'. apply binds_head.
         rewrite substAT_lclosed.
           apply weakeningDT_head. apply Ht. apply HD'.
           eapply regularityT. apply Ht.
    eapply ofE_gen.
      apply weakeningDG_head.      
        apply HG.
        apply okD_var.
         eapply validityDT; apply Ht.
         apply Fra.
      intros  b Frb.
      assert (okDa : okD (D & a ~ k )).
       apply okD_var.  eapply validityDT. apply Ht. apply Fra. 
      assert (okDab : okD (D & a ~ k & b ~ arrK k omK)).
       apply okD_var.  
       apply okDa. apply Frb.
      assert (HDab: ofT (D & a ~ k & b ~ arrK k omK) (appT b a) omK).
       eapply ofT_app.
        apply ofT_var. apply okDab. apply binds_head.
        apply weakeningDT_head. apply ofT_var.  apply okDa.
        apply binds_head.
        apply okDab.
      unfold openAE. simpl substAE.
      unfold openAT. simpl substAT.
      eapply ofE_lam.
        apply HDab.
        intros x Frx. unfold openXE. simpl substXE. apply ofE_var.
          apply okG_var. 
          apply weakeningDG_head.
           apply weakeningDG_head. 
            apply HG.  
            apply okD_var.
            apply (validityDT _ _ _ Ht).
            apply Fra.  
            apply okD_var.
            apply okD_var.
            apply (validityDT _ _ _ Ht). 
            apply Fra.
            apply Frb.
            apply Frx.
            eapply HDab.
            apply binds_head.
            apply Ht.
            (* *)
            eapply ofT_app.
            eapply ofT_lam.
            intros a Fra.
            unfold openAT. simpl substAT.
    apply ofT_all with (L := dom (D & a ~ k)). intros b Fr.
    unfold openAT.
    simpl substAT.
    assert (HD'' : okD (D & a ~ k )).
     apply okD_var.  eapply validityDT. apply Ht. apply Fra. 
    assert (HD' : okD (D & a ~ k & b ~ arrK k omK)).
      apply okD_var.  
      apply HD''. apply Fr.
    assert (Happ : ofT (D & a ~ k & b ~ arrK k omK) (appT b a) omK).
      eapply ofT_app.
        apply ofT_var. apply HD'. apply binds_head.
          apply weakeningDT_head. apply ofT_var.  apply HD''.
          apply binds_head.
    apply HD'.
    apply ofT_arr. apply Happ. apply Happ.
    apply Ht.
Qed.
