(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Metatheory_Misc.
Require Import AdditionalTactics.
Require Import Fomega_syntax.
Require Import Fomega_vars.
Require Import Fomega_typing.
Require Import Fomega_syntactic.


(*************************************************************************)
(* Regularity *)
(*************************************************************************)

Lemma regularityD : forall D,
  okD D ->
  ok D.
Proof.
  intros D HD.
  induction HD.
    Case "empty".
      apply ok_empty.
    Case "var".
      apply ok_push.
        apply IHHD.
        apply H.
Qed.

Lemma regularityG : forall G D,
  okG D G ->
  ok G.
Proof.
  intros G D HG.
  induction HG.
    Case "empty".
      apply ok_empty.
    Case "var".
      apply ok_push.
        apply IHHG.
        apply H.
Qed.

Lemma regularityT : forall D t k,
  ofT D t k ->
  lclosedT t.
Proof.
  intros D t k Ht.
  induction Ht; eauto*.
Qed.

Lemma regularityTT : forall t t',
  eqT t t' ->
  lclosedT t /\ lclosedT t'.
Proof.
  intros t t' Ht.
  induction Ht; eauto*; split;
    try solve
    [ first [apply lclosedT_rec with (L := L) | apply lclosedT_all with (L := L) |
             apply lclosedT_ex with (L := L) | apply lclosedT_lam with (L := L)];
        intros a Fr; unfold openAT; eapply proj1; apply (H0 a Fr)
    | first [apply lclosedT_rec with (L := L) | apply lclosedT_all with (L := L) |
             apply lclosedT_ex with (L := L) | apply lclosedT_lam with (L := L)];
        intros a Fr; unfold openAT; eapply proj2; apply (H0 a Fr)
    ].
    Case "lam_eta l".
      apply lclosedT_lam with (L := {}). intros a Fr.
      unfold openAT. simpl substAT.
      apply lclosedT_app.
        rewrite substAT_lclosed; apply H.
        apply lclosedT_var.
    Case "lam_eta r".
      apply H.
    Case "app_beta l".
      apply lclosedT_app.
        inversion H.
        apply lclosedT_lam with (L := L). intros a Fr.
          apply H2.
          apply Fr.
        apply H0.
    Case "app_beta r".
      inversion H.
      pick_fresh_gen (L \u varsAT t1) a.
      rewrite (openAT_as_substAT a) .
      apply  lclosedT_substAT.
      apply H0.
      apply H2.
      fresh_solve.
      fresh_solve.
Qed.

Lemma regularityE : forall D G e t,
  ofE D G e t ->
  lclosedE e.
Proof.
  intros D G e t He.
  induction He; eauto using regularityT.
Qed.


(*************************************************************************)
(* Environment validity *)
(*************************************************************************)

Lemma validityDD : forall D a k,
  okD (D & a~k) ->
  okD D.
Proof.
  intros D a k HD'.
  inversion HD'. apply H1.
Qed.

Lemma validityDD_split: forall D' D, okD (D & D') -> okD D.
Proof.
  induction D'.
  intros.
  fold (@empty knd) in H.
  rewrite concat_empty in H. auto.
  intros D H.
  assert (okD (D & D')).
  assert (D & (a :: D') = D & (D' & (a::empty))) as rw.
   reflexivity.
  rewrite rw in H.
  rewrite <- Env.concat_assoc in H.
  destruct a.
  apply (validityDD  _ _ _ H).
  eapply IHD'.
  auto.
Qed.


Lemma validityDT : forall D t k,
  ofT D t k ->
  okD D.
Proof.
  intros D t k Ht.
  induction Ht; auto.
    Case "rec".
      pick_fresh_gen L a.
      assert (HD := H0 a Fr). inversion HD.
      apply H3.
    Case "all".
      pick_fresh_gen L a.
      assert (HD := H0 a Fr). inversion HD.
      apply H3.
    Case "ex".
      pick_fresh_gen L a.
      assert (HD := H0 a Fr). inversion HD.
      apply H3.
    Case "lam".
      pick_fresh_gen L a.
      assert (HD := H0 a Fr). inversion HD.
      apply H3.
Qed.

Lemma validityGG : forall D G x t,
  okG D (G & x~t) ->
  okG D G.
Proof.
  intros D G x t HG'.
  inversion HG'. apply H3.
Qed.

Lemma validityDG : forall D G,
  okG D G ->
  okD D.
Proof.
  intros D G HG.
  induction HG; auto.
Qed.

Lemma validityTG : forall D G x t,
  okG D G ->
  binds x t G ->
  ofT D t omK.
Proof.
  intros D G x t HG Hx.
  induction HG.
    Case "nil".
      inversion Hx.
    Case "var".
      inversion Hx.
      destruct (x == x0).
        SCase "==".
          inversion H2.
          rewrite <- H3.
          apply H0.
        SCase "<>".
          apply IHHG.
          unfold binds.
          apply H2.
Qed.

Lemma validityGE : forall D G e t,
  ofE D G e t ->
  okG D G.
Proof.
  intros D G e t He.
  induction He; eauto.
    Case "lam".
      pick_fresh_gen L x.
      assert (HD := H1 x Fr).
      inversion HD.
      apply H6.
Qed.

Lemma validityDE : forall D G e t,
  ofE D G e t ->
  okD D.
Proof.
  intros D G e t He.
  eapply validityDG. eapply validityGE. apply He.
Qed.


(*************************************************************************)
(* Weakening *)
(*************************************************************************)

Lemma weakeningDA : forall D1 D D2 a k,
  binds a k (D1 & D2) ->
  okD (D1 & D & D2) ->
  binds a k (D1 & D & D2).
Proof.
  intros D1 D D2 a k. intros Ha HD'.
  apply binds_weaken.
    apply Ha.
    apply regularityD. apply HD'.
Qed.

Lemma weakeningDA_head : forall D1 D2 a k,
  binds a k D1 ->
  okD (D1 & D2) ->
  binds a k (D1 & D2).
Proof.
  intros D1 D2 a k. intros Ha HD.
  replace (D1 & D2) with (D1 & D2 & empty).
    apply weakeningDA.
      rewrite concat_empty. apply Ha.
      rewrite concat_empty. apply HD.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningGX : forall D G1 G G2 x t,
  binds x t (G1 & G2) ->
  okG D (G1 & G & G2) ->
  binds x t (G1 & G & G2).
Proof.
  intros D G1 G G2 x t. intros Hx HG'.
  apply binds_weaken.
    apply Hx.
    eapply regularityG. apply HG'.
Qed.

Lemma weakeningGX_head : forall D G1 G2 x t,
  binds x t G1 ->
  okG D (G1 & G2) ->
  binds x t (G1 & G2).
Proof.
  intros D G1 G2 x t. intros Hx HG.
  replace (G1 & G2) with (G1 & G2 & empty).
    eapply weakeningGX.
      rewrite concat_empty. apply Hx.
      rewrite concat_empty. apply HG.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningDT : forall D1 D D2 t k,
  ofT (D1 & D2) t k ->
  okD (D1 & D & D2) ->
  ofT (D1 & D & D2) t k.
Proof.
  intros D1 D D2 t k. intros Ht.
  remember (D1 & D2) as D12.
  generalize dependent D2.
  induction Ht; intros D2 Eq HD'; subst; eauto using (weakeningDA).
    Case "rec".
      apply ofT_rec with (L := L \u dom (D1 & D & D2)). intros a Fr.
      rewrite -> concat_assoc. apply H0.
        notin_solve.
        rewrite <- concat_assoc. reflexivity.
        rewrite <- concat_assoc. apply okD_var.
          apply HD'.
          notin_solve.
    Case "all".
      apply ofT_all with (L := L \u dom (D1 & D & D2)). intros a Fr.
      rewrite -> concat_assoc. apply H0.
        notin_solve.
        rewrite <- concat_assoc. reflexivity.
        rewrite <- concat_assoc. apply okD_var.
          apply HD'.
          notin_solve.
    Case "ex".
      apply ofT_ex with (L := L \u dom (D1 & D & D2)). intros a Fr.
      rewrite -> concat_assoc. apply H0.
        notin_solve.
        rewrite <- concat_assoc. reflexivity.
        rewrite <- concat_assoc. apply okD_var.
          apply HD'.
          notin_solve.
    Case "lam".
      apply ofT_lam with (L := L \u dom (D1 & D & D2)). intros a Fr.
      rewrite -> concat_assoc. apply H0.
        notin_solve.
        rewrite <- concat_assoc. reflexivity.
        rewrite <- concat_assoc. apply okD_var.
          apply HD'.
          notin_solve.
Qed.

Lemma weakeningDT_head : forall D1 D2 t k,
  ofT D1 t k ->
  okD (D1 & D2) ->
  ofT (D1 & D2) t k.
Proof.
  intros D1 D2 t k. intros Ht HD.
  replace (D1 & D2) with (D1 & D2 & empty).
    apply weakeningDT.
      rewrite concat_empty. apply Ht.
      rewrite concat_empty. apply HD.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningDG : forall D1 D D2 G,
  okG (D1 & D2) G ->
  okD (D1 & D & D2) ->
  okG (D1 & D & D2) G.
Proof.
  intros D1 D D2 G. intros HG.
  remember (D1 & D2) as D12.
  generalize dependent D2.
  induction HG; intros D2 Eq HD'; subst.
    Case "empty".
      apply okG_empty.
      apply HD'.
    Case "var".
      apply okG_var.
        apply IHHG. reflexivity. apply HD'.
        apply H.
        apply weakeningDT. apply H0. apply HD'.
Qed.

Lemma weakeningDG_head : forall D1 D2 G,
  okG D1 G ->
  okD (D1 & D2) ->
  okG (D1 & D2) G.
Proof.
  intros D1 D2 G. intros HG HD.
  replace (D1 & D2) with (D1 & D2 & empty).
    apply weakeningDG.
      rewrite concat_empty. apply HG.
      rewrite concat_empty. apply HD.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningDE : forall D1 D D2 G e t,
  ofE (D1 & D2) G e t ->
  okD (D1 & D & D2) ->
  ofE (D1 & D & D2) G e t.
Proof.
  intros D1 D D2 G e t. intros Ht.
  remember (D1 & D2) as D'.
  generalize dependent D2.
  induction Ht; intros D2 Eq HD'; subst; eauto using (weakeningDG).
    Case "lam".
      apply ofE_lam with (L := L \u dom (D1 & D & D2)).
        apply weakeningDT. apply H. apply HD'.
        intros x Fr. apply H1.
          notin_solve.
          reflexivity.
          apply HD'.
    Case "lft".
      apply ofE_lft.
        apply IHHt. reflexivity. apply HD'.
        apply weakeningDT. apply H. apply HD'.
    Case "rgt".
      apply ofE_rgt.
        apply IHHt. reflexivity. apply HD'.
        apply weakeningDT. apply H. apply HD'.
    Case "roll".
      apply ofE_roll.
        apply IHHt. reflexivity. apply HD'.
        apply weakeningDT. apply H. apply HD'.
    Case "gen".
      apply ofE_gen with (L := L \u dom (D1 & D & D2)).
        apply weakeningDG. apply H. apply HD'.
        intros a Fr. rewrite -> concat_assoc. apply H1.
          notin_solve.
          rewrite <- concat_assoc. reflexivity.
          rewrite <- concat_assoc. apply okD_var.
            apply HD'.
            notin_solve.
    Case "inst".
      eapply ofE_inst.
        apply IHHt. reflexivity. apply HD'.
        apply weakeningDT. apply H. apply HD'.
    Case "pack".
      eapply ofE_pack.
        apply weakeningDT. apply H. apply HD'.
        apply IHHt. reflexivity. apply HD'.
        apply weakeningDT. apply H0.
        apply HD'.
    Case "unpack".
      eapply ofE_unpack with (L := L \u dom (D1 & D & D2)).
        apply IHHt. reflexivity. apply HD'.
        intros a x FrA FrX. rewrite -> concat_assoc. apply H0.
          notin_solve. notin_solve.
          rewrite <- concat_assoc. reflexivity.
          rewrite <- concat_assoc. apply okD_var.
            apply HD'.
            notin_solve.
        apply weakeningDT. apply H1. apply HD'.
    Case "sub".
      eapply ofE_sub.
        apply IHHt. reflexivity. apply HD'.
        apply weakeningDT. apply H. apply HD'.
        apply H0.
Qed.

Lemma weakeningDE_head : forall D1 D2 G e t,
  ofE D1 G e t ->
  okD (D1 & D2) ->
  ofE (D1 & D2) G e t.
Proof.
  intros D1 D2 G e t. intros He HD.
  replace (D1 & D2) with (D1 & D2 & empty).
    apply weakeningDE.
      rewrite concat_empty. apply He.
      rewrite concat_empty. apply HD.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningGE : forall D G1 G G2 e t,
  ofE D (G1 & G2) e t ->
  okG D (G1 & G & G2) ->
  ofE D (G1 & G & G2) e t.
Proof.
  intros D G1 G G2 e t. intros He.
  remember (G1 & G2) as G'.
  generalize dependent G2.
  induction He; intros G2 Eq HG'; subst; eauto using (weakeningGX).
    Case "lam".
      apply ofE_lam with (L := L \u dom (G1 & G & G2)).
        apply H.
        intros x Fr. rewrite -> concat_assoc. apply H1.
          notin_solve.
          rewrite <- concat_assoc. reflexivity.
          rewrite <- concat_assoc. apply okG_var.
            apply HG'.
            notin_solve.
            apply H.
    Case "gen".
      apply ofE_gen with (L := L \u dom D).
        apply HG'.
        intros x Fr. apply H1.
          notin_solve.
          reflexivity.
          replace (D & x~k) with (D & x~k & empty).
            apply weakeningDG. apply HG'.
            rewrite -> concat_empty. apply okD_var.
              eapply validityDG. apply H.
              notin_solve.
              reflexivity. 
    Case "unpack".
      eapply ofE_unpack with (L := L \u dom D \u dom (G1 & G & G2)).
        apply IHHe. reflexivity. apply HG'.
        intros a x FrA FrX. rewrite -> concat_assoc. apply H0.
          notin_solve. notin_solve.
          reflexivity.
          replace (D & a~k1) with (D & a~k1 & empty).
            rewrite <- concat_assoc. apply okG_var.
              apply weakeningDG.
                apply HG'.
                rewrite -> concat_empty. apply okD_var.
                  eapply validityDG. apply HG'.
                  notin_solve.
              notin_solve.
              (**)
                assert (HG : okG (D & a~k1) (G1 & G2 & x~(openAT a t1))).
                  eapply validityGE. apply H.
                    notin_solve. notin_solve.
                eapply validityTG.
                  apply HG.
                  apply binds_head.
            reflexivity.
            apply H1.
Qed.

Lemma weakeningGE_head : forall D G1 G2 e t,
  ofE D G1 e t ->
  okG D (G1 & G2) ->
  ofE D (G1 & G2) e t.
Proof.
  intros D G1 G2 e t. intros He HG.
  replace (G1 & G2) with (G1 & G2 & empty).
    eapply weakeningGE.
      rewrite concat_empty. apply He.
      rewrite concat_empty. apply HG.
    rewrite concat_empty. reflexivity.
Qed.


(*************************************************************************)
(* Well-formedness of types under type substitution *)
(*************************************************************************)

Lemma substitutabilityAT : forall D1 a D2 t' k' t k,
  ofT (D1 & a~k' & D2) t k ->
  ofT (D1 & D2) t' k' ->
  ofT (D1 & D2) (substAT a t' t) k.
Proof.
  intros D1 a D2 t' k' t k Ht Ht'.
  remember (D1 & a~k' & D2) as D.
  generalize dependent D2.
  induction Ht; intros D2 Ht' Eq; subst; simpl substAT.
    Case "var".
      destruct (a0 == a); subst.
        SCase "a0 = a".
          assert (Eq : k = k').
            eapply binds_mid_eq. apply H0. apply regularityD. apply H.
          subst. apply Ht'.
        SCase "a0 <> a".
          apply ofT_var.
            eapply validityDT. apply Ht'.
            binds_cases H0.
              apply binds_concat_ok. apply B0. eapply ok_remove. eapply regularityD. apply H.
              apply binds_prepend. apply B0.
    Case "unit".
      apply ofT_unit.
        eapply validityDT. apply Ht'.
    Case "arr".
      apply ofT_arr.
        apply IHHt1. apply Ht'. reflexivity.
        apply IHHt2. apply Ht'. reflexivity.
    Case "tim".
      apply ofT_tim.
        apply IHHt1. apply Ht'. reflexivity.
        apply IHHt2. apply Ht'. reflexivity.
    Case "plus".
      apply ofT_plus.
        apply IHHt1. apply Ht'. reflexivity.
        apply IHHt2. apply Ht'. reflexivity.
    Case "rec".
      apply ofT_rec with (L := L \u dom (D1 & D2) \u {{a}}).
      intros a0 Fr.
      rewrite -> concat_assoc.
      rewrite openAT_substAT.
        apply H0.
          notin_solve.
          rewrite <- concat_assoc.
            apply weakeningDT_head.
              apply Ht'.
              apply okD_var. eapply validityDT. apply Ht'. notin_solve.
          apply concat_assoc.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        eapply regularityT. apply Ht'.
    Case "all".
      apply ofT_all with (L := L \u dom (D1 & D2) \u {{a}}).
      intros a0 Fr.
      rewrite -> concat_assoc.
      rewrite openAT_substAT.
        apply H0.
          notin_solve.
          rewrite <- concat_assoc.
            apply weakeningDT_head.
              apply Ht'.
              apply okD_var. eapply validityDT. apply Ht'. notin_solve.
          apply concat_assoc.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        eapply regularityT. apply Ht'.
    Case "ex".
      apply ofT_ex with (L := L \u dom (D1 & D2) \u {{a}}).
      intros a0 Fr.
      rewrite -> concat_assoc.
      rewrite openAT_substAT.
        apply H0.
          notin_solve.
          rewrite <- concat_assoc.
            apply weakeningDT_head.
              apply Ht'.
              apply okD_var. eapply validityDT. apply Ht'. notin_solve.
          apply concat_assoc.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        eapply regularityT. apply Ht'.
    Case "triv".
      apply ofT_triv.
        eapply validityDT. apply Ht'.
    Case "lam".
      apply ofT_lam with (L := L \u dom (D1 & D2) \u {{a}}).
      intros a0 Fr.
      rewrite -> concat_assoc.
      rewrite openAT_substAT.
        apply H0.
          notin_solve.
          rewrite <- concat_assoc.
            apply weakeningDT_head.
              apply Ht'.
              apply okD_var. eapply validityDT. apply Ht'. notin_solve.
          apply concat_assoc.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        eapply regularityT. apply Ht'.
    Case "app".
      eapply ofT_app.
        apply IHHt1. apply Ht'. reflexivity.
        apply IHHt2. apply Ht'. reflexivity.
    Case "pair".
      apply ofT_pair.
        apply IHHt1. apply Ht'. reflexivity.
        apply IHHt2. apply Ht'. reflexivity.
    Case "fst".
      eapply ofT_fst.
        apply IHHt. apply Ht'. reflexivity.
    Case "snd".
      eapply ofT_snd.
        apply IHHt. apply Ht'. reflexivity.
Qed.

Lemma substitutabilityAT_head : forall D a t' k' t k,
  ofT (D & a~k') t k ->
  ofT D t' k' ->
  ofT D (substAT a t' t) k.
Proof.
  intros D a t' k' t k Ht Ht'.
  replace D with (D & empty).
    eapply substitutabilityAT.
      rewrite concat_empty. apply Ht.
      rewrite concat_empty. apply Ht'.
    rewrite concat_empty. reflexivity.
Qed.

Lemma openingAT : forall D t' k' t k L,
  (forall a, a \notin L -> ofT (D & a~k') (openAT a t) k) ->
  ofT D t' k' ->
  ofT D (openAT t' t) k.
Proof.
  intros D t' k' t k L Ht Ht'.
  pick_fresh_gen (L \u varsAT t) a.
  assert (Fr' : a \notin L). notin_solve.
  assert (Ht'' := Ht a Fr').
  rewrite openAT_as_substAT with (a := a).
    replace D with (D & empty).
      eapply substitutabilityAT.
        rewrite -> concat_empty. apply Ht''.
      apply Ht'.
    apply concat_empty.
  notin_solve.
Qed.


(*************************************************************************)
(* Equivalence of types under type substitution *)
(*************************************************************************)

Lemma substitutabilityATT : forall (a : var) t' t1 t2,
  eqT t1 t2 ->
  lclosedT t' ->
  eqT (substAT a t' t1) (substAT a t' t2).
Proof.
  intros a t' t1 t2 Heq Ht'.
  induction Heq; simpl substAT.
    Case "var".
      destruct (a0 == a); subst.
        apply eqT_refl. apply Ht'.
        apply eqT_var.
    Case "unit".
      apply eqT_unit.
    Case "arr".
      apply eqT_arr. apply IHHeq1. apply IHHeq2.
    Case "tim".
      apply eqT_tim. apply IHHeq1. apply IHHeq2.
    Case "plus".
      apply eqT_plus. apply IHHeq1. apply IHHeq2.
    Case "rec".
      apply eqT_rec with (L := L \u {{a}}).
      intros a0 Fr.
      rewrite openAT_substAT. rewrite openAT_substAT.
        apply H0. notin_solve.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
    Case "all".
      apply eqT_all with (L := L \u {{a}}).
      intros a0 Fr.
      rewrite openAT_substAT. rewrite openAT_substAT.
        apply H0. notin_solve.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
    Case "ex".
      apply eqT_ex with (L := L \u {{a}}).
      intros a0 Fr.
      rewrite openAT_substAT. rewrite openAT_substAT.
        apply H0. notin_solve.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
    Case "triv".
      apply eqT_triv.
    Case "lam".
      apply eqT_lam with (L := L \u {{a}}).
      intros a0 Fr.
      rewrite openAT_substAT. rewrite openAT_substAT.
        apply H0. notin_solve.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
        intro Eq. inversion Eq.
        unfold indepAT. simpl varsAT. notin_solve.
        apply Ht'.
    Case "lam_eta".
      apply eqT_lam_eta. apply lclosedT_substAT. apply Ht'. apply H.
    Case "app".
      apply eqT_app. apply IHHeq1. apply IHHeq2.
    Case "app_beta".
      unfold openAT. rewrite substAT_substAT_dep.
        apply eqT_app_beta.
          change ( lclosedT (substAT a t' (lamT k t1))).
          apply lclosedT_substAT. apply Ht'. apply H.
          apply lclosedT_substAT. apply Ht'. apply H0.
        intro Eq. inversion Eq.
        unfold indepAT. apply Ht'.
    Case "pair".
      apply eqT_pair. apply IHHeq1. apply IHHeq2.
    Case "pair_eta".
      apply eqT_pair_eta. apply lclosedT_substAT. apply Ht'. apply H.
    Case "fst".
      apply eqT_fst. apply IHHeq.
    Case "fst_beta".
      apply eqT_fst_beta.
        apply lclosedT_substAT. apply Ht'. apply H.
        apply lclosedT_substAT. apply Ht'. apply H0.
    Case "snd".
      apply eqT_snd. apply IHHeq.
    Case "snd_beta".
      apply eqT_snd_beta.
        apply lclosedT_substAT. apply Ht'. apply H.
        apply lclosedT_substAT. apply Ht'. apply H0.
    Case "refl".
      apply eqT_refl. apply lclosedT_substAT. apply Ht'. apply H.
    Case "sym".
      apply eqT_sym. apply IHHeq.
    Case "trans".
      eapply eqT_trans. apply IHHeq1. apply IHHeq2.
Qed.


(*************************************************************************)
(* Validity of types *)
(*************************************************************************)

Lemma validityTE : forall D G e t,
  ofE D G e t ->
  ofT D t omK.
Proof.
  intros D G e t He.
  induction He; eauto.
    Case "var".
      eapply validityTG.
      apply H. apply H0.
    Case "triv".
      apply ofT_unit.
        eapply validityDG. apply H.
    Case "lam".
      pick_fresh_gen L x.
      apply ofT_arr.
        apply H.
        eapply H1. apply Fr.
    Case "app".
      inversion IHHe1. apply H3.
    Case "fst".
      inversion IHHe. apply H2.
    Case "snd".
      inversion IHHe. apply H3.
    Case "unroll".
      inversion IHHe; subst.
      eapply openingAT.
        apply H3.
        apply IHHe.
    Case "inst".
      inversion IHHe; subst.
      eapply openingAT.
        apply H2.
        apply H.
Qed.


(*************************************************************************)
(* Well-formedness of term environments under type substitution *)
(*************************************************************************)

Lemma substitutabilityAG : forall D1 a D2 t' k' G,
  okG (D1 & a~k' & D2) G ->
  ofT (D1 & D2) t' k' ->
  okG (D1 & D2) (substAG a t' G).
Proof.
  intros D1 a D2 t' k' G HG Ht'.
  remember (D1 & a~k' & D2) as D.
  generalize dependent D2.
  induction HG; intros D2 Ht' Eq; subst; unfold substAG; simpl Env.map.
    Case "empty".
      apply okG_empty.
        eapply validityDT. apply Ht'.
    Case "var".
      assert (Aux : forall x (t :typ) G, (x,t)::G = G & x~t).
        intros. reflexivity.
      rewrite Aux.
      apply okG_var.
        apply IHHG. apply Ht'. reflexivity.
        rewrite dom_map. apply H.
        eapply substitutabilityAT. apply H0. apply Ht'.
Qed.


(*************************************************************************)
(* Well-formedness of terms under type substitution *)
(*************************************************************************)

Lemma substitutabilityAE : forall D1 a D2 t' k' G e t,
  ofE (D1 & a~k' & D2) G e t ->
  ofT (D1 & D2) t' k' ->
  ofE (D1 & D2) (substAG a t' G) (substAE a t' e) (substAT a t' t).
Proof.
  intros D1 a D2 t' k' G e t He Ht'.
  remember (D1 & a~k' & D2) as D.
  generalize dependent D2.
  induction He; intros D2 Ht' Eq; subst; simpl substAT; simpl substAE.
    Case "var".
      apply ofE_var.
        eapply substitutabilityAG. apply H. apply Ht'.
        unfold substAG. apply binds_map. apply H0.
    Case "triv".
      apply ofE_triv.
        eapply substitutabilityAG. apply H. apply Ht'.
    Case "lam".
      apply ofE_lam with (L := L).
        eapply substitutabilityAT. apply H. apply Ht'.
        intros x Fr.
          unfold substAG. rewrite <- map_concat_single. fold (substAG a t' (G & x~t1)).
          rewrite openXE_substAE.
            apply H1. notin_solve. apply Ht'. reflexivity.
            unfold indepAE. simpl varsAE. notin_solve.
    Case "app".
      eapply ofE_app.
        simpl substAT in IHHe1. apply IHHe1. apply Ht'. reflexivity.
        apply IHHe2. apply Ht'. reflexivity.
    Case "pair".
      apply ofE_pair.
        apply IHHe1. apply Ht'. reflexivity.
        apply IHHe2. apply Ht'. reflexivity.
    Case "fst".
      eapply ofE_fst.
        simpl substAT in IHHe. apply IHHe. apply Ht'. reflexivity.
    Case "snd".
      eapply ofE_snd.
        simpl substAT in IHHe. apply IHHe. apply Ht'. reflexivity.
    Case "lft".
      eapply ofE_lft.
        apply IHHe. apply Ht'. reflexivity.
        replace (plusT (substAT a t' t1) (substAT a t' t2)) with (substAT a t' (plusT t1 t2)).
          eapply substitutabilityAT. apply H. apply Ht'.
          simpl substAT. reflexivity.
    Case "rgt".
      eapply ofE_rgt.
        apply IHHe. apply Ht'. reflexivity.
        replace (plusT (substAT a t' t1) (substAT a t' t2)) with (substAT a t' (plusT t1 t2)).
          eapply substitutabilityAT. apply H. apply Ht'.
          simpl substAT. reflexivity.
    Case "roll".
      eapply ofE_roll.
        unfold openAT in IHHe. rewrite substAT_substAT_dep in IHHe.
          apply IHHe. apply Ht'. reflexivity.
          intro Eq. inversion Eq.
          unfold indepAT. eapply regularityT. apply Ht'.
        replace (recT k (substAT a t' t)) with (substAT a t' (recT k t)).
          eapply substitutabilityAT. apply H. apply Ht'.
          simpl substAT. reflexivity.
    Case "unroll".
      unfold openAT. rewrite substAT_substAT_dep. simpl substAT.
        eapply ofE_unroll.
          simpl substAT in IHHe. apply IHHe. apply Ht'. reflexivity.
        intro Eq. inversion Eq.
        unfold indepAT. eapply regularityT. apply Ht'.
    Case "gen".
      apply ofE_gen with (L := L \u dom (D1 & D2) \u {{a}}).
        eapply substitutabilityAG. apply H. apply Ht'.
        intros a0 Fr.
          rewrite -> concat_assoc.
          rewrite openAT_substAT.
          rewrite openAE_substAE.
          apply H1.
            notin_solve.
            rewrite <- concat_assoc. apply weakeningDT_head.
              apply Ht'.
              apply okD_var. eapply validityDT. apply Ht'. notin_solve.
            apply concat_assoc.
          intro Eq. inversion Eq.
          unfold indepAT. simpl varsAT. notin_solve.
          eapply regularityT. apply Ht'.
          intro Eq. inversion Eq.
          unfold indepAT. simpl varsAT. notin_solve.
          eapply regularityT. apply Ht'.
    Case "inst".
      unfold openAT. rewrite substAT_substAT_dep.
        eapply ofE_inst.
          simpl substAT in IHHe. apply IHHe. apply Ht'. reflexivity.
          eapply substitutabilityAT. apply H. apply Ht'.
        intro Eq. inversion Eq.
        unfold indepAT. eapply regularityT. apply Ht'.
    Case "pack".
      eapply ofE_pack.
        eapply substitutabilityAT. apply H. apply Ht'.
        unfold openAT in IHHe. rewrite substAT_substAT_dep in IHHe.
          apply IHHe. apply Ht'. reflexivity.
          intro Eq. inversion Eq.
          unfold indepAT. eapply regularityT. apply Ht'.
        replace (exT k1 (substAT a t' t2)) with (substAT a t' (exT k1 t2)).
          eapply substitutabilityAT. apply H0. apply Ht'.
          simpl substAT. reflexivity.
    Case "unpack".
      eapply ofE_unpack with (L := L \u dom (D1 & D2) \u {{a}}).
        simpl substAT in IHHe. apply IHHe. apply Ht'. reflexivity.
        intros a0 x Fra Frx.
          rewrite -> concat_assoc.
          rewrite openAT_substAT.
          unfold substAG. rewrite <- map_concat_single. fold (substAG a t').
          rewrite openAE_substAE.
          rewrite openXE_substAE.
          apply H0.
            notin_solve.
            notin_solve.
            rewrite <- concat_assoc. apply weakeningDT_head.
              apply Ht'.
              apply okD_var. eapply validityDT. apply Ht'. notin_solve.
            apply concat_assoc.
          unfold indepAE. simpl varsAE. notin_solve.
          intro Eq. inversion Eq.
          unfold indepAT. simpl varsAT. notin_solve.
          eapply regularityT. apply Ht'.
          intro Eq. inversion Eq.
          unfold indepAT. simpl varsAT. notin_solve.
          eapply regularityT. apply Ht'.
        eapply substitutabilityAT. apply H1. apply Ht'.
    Case "sub".
      apply ofE_sub with (t' := substAT a t' t'0).
        apply IHHe. apply Ht'. reflexivity.
        eapply substitutabilityAT. apply H. apply Ht'.
        eapply substitutabilityATT. apply H0. eapply regularityT. apply Ht'.
Qed.

Lemma substitutabilityAE_head : forall D a t' k' G e t,
  ofE (D & a~k') G e t ->
  ofT D t' k' ->
  ofE D (substAG a t' G) (substAE a t' e) (substAT a t' t).
Proof.
  intros D a t' k' G e t He Ht'.
  replace D with (D & empty).
    eapply substitutabilityAE.
      rewrite concat_empty. apply He.
      rewrite concat_empty. apply Ht'.
    rewrite concat_empty. reflexivity.
Qed.

Lemma openingAE : forall D t' k' G e t L,
  okG D G ->
  (forall a, a \notin L -> ofE (D & a~k') G (openAE a e) (openAT a t)) ->
  ofT D t' k' ->
  ofE D G (openAE t' e) (openAT t' t).
Proof.
  intros D t' k' G e t L HG He Ht'.
  pick_fresh_gen (L \u varsAG G \u varsAT t \u varsAE e) a.
  assert (Fr' : a \notin L). notin_solve.
  assert (He' := He a Fr').
  rewrite openAT_as_substAT with (a := a).
  rewrite openAE_as_substAE with (a := a).
  replace D with (D & empty).
  replace G with (substAG a t' G).
    eapply substitutabilityAE.
      rewrite -> concat_empty. apply He'.
      apply Ht'.
    apply substAG_fresh. notin_solve.
    apply concat_empty.
    notin_solve.
    notin_solve.
Qed.


(*************************************************************************)
(* Well-formedness of terms under term substitution *)
(*************************************************************************)

Lemma substitutabilityXE : forall D G1 x G2 e' t' e t,
  ofE D (G1 & x~t' & G2) e t ->
  ofE D (G1 & G2) e' t' ->
  ofE D (G1 & G2) (substXE x e' e) t.
Proof.
  intros D G1 x G2 e' t' e t He He'.
  remember (G1 & x~t' & G2) as G.
  generalize dependent G2.
  induction He; intros G2 He' Eq; subst; simpl substXE.
    Case "var".
      destruct (x0 == x); subst.
        SCase "x0 = x".
          assert (Eq : t = t').
            eapply binds_mid_eq. apply H0. eapply regularityG. apply H.
          subst. apply He'.
        SCase "x0 <> x".
          apply ofE_var.
            eapply validityGE. apply He'.
            binds_cases H0.
              apply binds_concat_ok. apply B0. eapply ok_remove. eapply regularityG. apply H.
              apply binds_prepend. apply B0.
    Case "triv".
      apply ofE_triv.
        eapply validityGE. apply He'.
    Case "lam".
      apply ofE_lam with (L := L \u dom (G1 & G2) \u {{x}}).
      apply H.
      intros x0 Fr.
      rewrite -> concat_assoc.
      rewrite openXE_substXE.
        apply H1.
          notin_solve.
          rewrite <- concat_assoc.
            apply weakeningGE_head.
              apply He'.
              apply okG_var. eapply validityGE. apply He'. notin_solve. apply H.
          apply concat_assoc.
        intro Eq. inversion Eq.
        unfold indepXE. simpl varsXE. notin_solve.
        eapply regularityE. apply He'.
    Case "app".
      eapply ofE_app.
        apply IHHe1. apply He'. reflexivity.
        apply IHHe2. apply He'. reflexivity.
    Case "pair".
      apply ofE_pair.
        apply IHHe1. apply He'. reflexivity.
        apply IHHe2. apply He'. reflexivity.
    Case "fst".
      eapply ofE_fst.
        apply IHHe. apply He'. reflexivity.
    Case "snd".
      eapply ofE_snd.
        apply IHHe. apply He'. reflexivity.
    Case "lft".
      eapply ofE_lft.
        apply IHHe. apply He'. reflexivity.
        apply H.
    Case "rgt".
      eapply ofE_rgt.
        apply IHHe. apply He'. reflexivity.
        apply H.
    Case "roll".
      eapply ofE_roll.
        apply IHHe. apply He'. reflexivity.
        apply H.
    Case "unroll".
      eapply ofE_unroll.
        apply IHHe. apply He'. reflexivity.
    Case "gen".
      apply ofE_gen with (L := L \u dom D).
        eapply validityGE. apply He'.
        intros a0 Fr.
          rewrite openAE_substXE.
          apply H1.
            notin_solve.
            apply weakeningDE_head.
              apply He'.
              apply okD_var. eapply validityDE. apply He'. notin_solve.
            reflexivity.
          eapply regularityE. apply He'.
    Case "inst".
      eapply ofE_inst.
        apply IHHe. apply He'. reflexivity.
        apply H.
    Case "pack".
      eapply ofE_pack.
        apply H.
        apply IHHe. apply He'. reflexivity.
        apply H0.
    Case "unpack".
      assert (Ht1 : ofT D (exT k1 t1) omK). eapply validityTE. apply He.
      inversion Ht1.
      eapply ofE_unpack with (L := L \u L0 \u dom D \u dom (G1 & G2) \u {{x}}).
        apply IHHe. apply He'. reflexivity.
        intros a0 x0 Fra Frx.
          rewrite -> concat_assoc.
          rewrite openAE_substXE.
          rewrite openXE_substXE.
          apply H0.
            notin_solve.
            notin_solve.
            rewrite <- concat_assoc. apply weakeningGE_head. apply weakeningDE_head.
              apply He'.
              apply okD_var. eapply validityDT. apply H1. notin_solve.
              apply okG_var. apply weakeningDG_head.
                eapply validityGE. apply He'.
                  apply okD_var. eapply validityDT. apply H1. notin_solve.
                  notin_solve.
              apply H4. notin_solve.
            apply concat_assoc.
          intro Eq. inversion Eq.
          unfold indepXE. simpl varsXE. notin_solve.
          eapply regularityE. apply He'.
          eapply regularityE. apply He'.
          apply H1.
    Case "sub".
      apply ofE_sub with (t' := t'0).
        apply IHHe. apply He'. reflexivity.
        apply H.
        apply H0.
Qed.

Lemma substitutabilityXE_head : forall D G x e' t' e t,
  ofE D (G & x~t') e t ->
  ofE D G e' t' ->
  ofE D G (substXE x e' e) t.
Proof.
  intros D G x e' t' e t He He'.
  replace G with (G & empty).
    eapply substitutabilityXE.
      rewrite concat_empty. apply He.
      rewrite concat_empty. apply He'.
    rewrite concat_empty. reflexivity.
Qed.

Lemma openingXE : forall D G e' t' e t L,
  okG D G ->
  (forall x, x \notin L -> ofE D (G & x~t') (openXE x e) t) ->
  ofE D G e' t' ->
  ofE D G (openXE e' e) t.
Proof.
  intros D G e' t' e t L HG He He'.
  pick_fresh_gen (L \u varsXE e) x.
  assert (Fr' : x \notin L). notin_solve.
  assert (He'' := He x Fr').
  rewrite openXE_as_substXE with (x := x).
  replace G with (G & empty).
    eapply substitutabilityXE.
      rewrite -> concat_empty. apply He''.
      apply He'.
    apply concat_empty.
    notin_solve.
Qed.
